// Copyright (c) 2010 Dirk Jagdmann <doj@cubic.org>

// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any damages
// arising from the use of this software.

// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it
// freely, subject to the following restrictions:

//     1. The origin of this software must not be misrepresented; you
//        must not claim that you wrote the original software. If you use
//        this software in a product, an acknowledgment in the product
//        documentation would be appreciated but is not required.

//     2. Altered source versions must be plainly marked as such, and
//        must not be misrepresented as being the original software.

//     3. This notice may not be removed or altered from any source
//        distribution.

// $Header: /code/c++int/c++int.cc,v 1.2 2010/11/08 06:15:07 doj Exp $

#include <cstdio>
#include <cstdlib>
#include <errno.h>
#include <iostream>
#include <string>
#include <sysexits.h>
#include "doj/path.hpp"
#include "doj/file.hpp"
#include "md5.h"

const char *CXX_INT_DEBUG;

class TmpFile
{
protected:
  std::string fn_;
public:
  TmpFile(const std::string& filename) :
    fn_(filename)
  { }

  ~TmpFile()
  {
    if(!fn_.empty())
      {
	if(CXX_INT_DEBUG)
	  std::clog << "unlink " << fn_ << std::endl;
	unlink(fn_.c_str());
      }
  }

  const std::string& fn() const { return fn_; }

  TmpFile& operator= (const std::string& new_fn)
  {
    if(!fn_.empty())
      unlink(fn_.c_str());
    fn_=new_fn;
    return *this;
  }

  TmpFile& swap(TmpFile& tf)
  {
    fn_.swap(tf.fn_);
    return *this;
  }

private:
  TmpFile();
  TmpFile(const TmpFile&);
  TmpFile& operator= (const TmpFile&);
};

std::ostream& operator<< (std::ostream& os, const TmpFile& tf)
{
  return os << tf.fn();
}

class TmpFileExt : public TmpFile
{
public:
  TmpFileExt(const std::string& extension) :
    TmpFile("")
  {
    fn_=tmpnam(NULL);
    fn_+='.';
    fn_+=extension;
  }
};

std::string source_fn;

int runProgram(const std::string& prg, char **argv, int begin, int end)
{
  std::string cmd = prg;
  for(int i=begin; i<end; ++i)
    {
      cmd+=' ';
      cmd+=argv[i];
    }

  if(CXX_INT_DEBUG)
      std::clog << "execute: " << cmd << std::endl;

  const int status=system(cmd.c_str());
  if(status<0)
    {
      std::cerr << "could not execute '" << cmd << "' : " << strerror(errno) << std::endl;
      return EX_OSERR;
    }

  const int ret=WEXITSTATUS(status);
  if(ret != 0)
      std::cerr << "'" << cmd << "' exit status: " << ret << std::endl;

  return ret;
}

int main(int argc, char **argv)
{
  int compile_argc_begin=-1, compile_argc_end=-1, run_argc_begin=-1, run_argc_end=-1;

  CXX_INT_DEBUG=getenv("CXX_INT_DEBUG");

  if(CXX_INT_DEBUG)
  {
      for(int i=0; i<argc; ++i)
	  std::clog << "argv[" << i << "]=" << argv[i] << std::endl;
  }

  // check for -- argument
  {
      std::string a="--";
      for(int i=1; i<argc; ++i)
      {
	  if(a == argv[i])
	  {
	      compile_argc_begin=1;
	      compile_argc_end=i;
	      run_argc_begin=i+1;
	      run_argc_end=argc;
	      break;
	  }
      }
      // no -- was found
      if(compile_argc_begin!=1)
      {
	  compile_argc_begin=1;
	  compile_argc_end=argc;
      }
  }

  if((argc < 2) || (compile_argc_begin>=compile_argc_end))
    {
      std::cerr << "usage: c++int [c++ arguments] <source filename> -- [run arguments]\n"
	"environment variables:\n"
	"CXX: c++ compiler to use\n"
	"CXX_INT_CACHE: directory name to store compiled binaries\n"
	"CXX_INT_DEBUG: output some debug information\n"
	;
      return EX_USAGE;
    }

  // last argument is the c++ source filename
  const int source_argc = compile_argc_end-1;
  source_fn=argv[source_argc];
  if(!doj::exists(source_fn))
    {
      std::cerr << source_fn << " not found." << std::endl;
      return EX_NOINPUT;
    }

  std::string exe;

  // check if cache dir exists
  const char *CXX_INT_CACHE=getenv("CXX_INT_CACHE");
  if(CXX_INT_CACHE && doj::isDir(CXX_INT_CACHE))
    {
      // construct binary name
      std::string cache_fn=CXX_INT_CACHE;
      cache_fn+='/';
      cache_fn+=MD5Data(doj::absFilename(source_fn));
      // check if binary exists
      if(doj::isExec(cache_fn))
	{
	  // check if binary is newer than source
	  const time_t src_t=doj::mTime(source_fn);
	  const time_t exe_t=doj::mTime(cache_fn);
	  if(exe_t > src_t)
	    {
	      if(CXX_INT_DEBUG)
		std::clog << "using cached binary: " << cache_fn << std::endl;
	      return runProgram(cache_fn, argv, run_argc_begin, run_argc_end);
	    }
	}
      // store new compiled binary in cache dir
      exe=cache_fn;
    }

  TmpFileExt src_fn("cpp");
  TmpFile exe_(src_fn.fn() + ".exe");

  // if exe is empty, we did not branch into the if(CXX_INT_CACHE) above
  if(exe.empty())
    exe=exe_.fn();

  // copy source to modify she-bang
  {
    doj::FILEp in(source_fn, "rb");
    if(!in)
      {
	std::cerr << "could not open " << source_fn << " : " << strerror(errno) << std::endl;
	return EX_NOINPUT;
      }
    // skip first two characters
    in.getc();
    in.getc();

    doj::FILEp out(src_fn.fn(), "wb");
    if(!out)
      {
	std::cerr << "could not open " << src_fn << " : " << strerror(errno) << std::endl;
	return EX_CANTCREAT;
      }
    // write comment as first two characters
    out.putc('/');
    out.putc('/');

    while(true)
      {
	char buf[65536];
	int r=in.read(buf, sizeof(buf));
	if(r < 0)
	  {
	    std::cerr << "could not read from " << source_fn << " : " << strerror(errno) << std::endl;
	    return EX_IOERR;
	  }
	if(r == 0)
	  break;
	int w=out.write(buf, r);
	if(w != r)
	  {
	    std::cerr << "could not write to " << source_fn << " : " << strerror(errno) << std::endl;
	    return EX_IOERR;
	  }
      }

    argv[source_argc]=strdup(src_fn.fn().c_str()); // we ignore this memory leak
  }

  // compile
  const char *CXX=getenv("CXX");
  std::string cmd = CXX ? CXX : "c++";
  cmd+=" -o ";
  cmd+=exe;

  if(runProgram(cmd, argv, compile_argc_begin, compile_argc_end) != 0)
      return EX_OSERR;
  if(!doj::isExec(exe))
    {
      std::cerr << "could not locate compiled binary " << exe << std::endl;
      return EX_DATAERR;
    }

  // run
  return runProgram(exe, argv, run_argc_begin, run_argc_end);
}
