/* Copyright (c) 2004..2008 Dirk Jagdmann <doj@cubic.org>

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you
       must not claim that you wrote the original software. If you use
       this software in a product, an acknowledgment in the product
       documentation would be appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and
       must not be misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
       distribution. */

// $Header: /code/doj/path.cpp,v 1.54 2011/05/07 14:33:13 doj Exp $

//lint -save -e1923

#ifdef _MSC_VER
#define _CRT_SECURE_NO_WARNINGS 1
#include <direct.h>
#include <io.h>
#define PATH_MAX MAX_PATH
#define snprintf _snprintf
#define DIRSEP_C '\\'
#define DIRSEP_S "\\"

#define syslog(...)

#pragma warning (disable: 4800 4996)

#endif

#include <errno.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>

#if defined(__unix__) || defined(__APPLE__)
#include <syslog.h>
#include <dirent.h>
#include <pwd.h>
#include <unistd.h>
#define DIRSEP_C '/'
#define DIRSEP_S "/"
#endif

#include <iostream>
#include <fstream>
#include <limits>

#include "path.hpp"
#include "readn.h"

#ifndef DIRSEP_C
#error directory separator character not defined on your platform!
#endif
#ifndef DIRSEP_S
#error directory separator character (as string) not defined!
#endif

#ifdef __APPLE__
#define __unix__ 1
#endif

std::string doj::cleanupPath(std::string p)
{
  std::string::size_type pos=0;

  // \todo replace ~user on win32
#ifdef __unix__
  // replace ~user
  while(pos!=std::string::npos)
    {
      pos=p.find("~");
      if(pos!=std::string::npos)
	{
	  std::string u;
	  std::string::size_type Pos=pos+1;
	  while(p[Pos] != DIRSEP_C && Pos < p.size())
	    u+=p[Pos++];
	  if(!u.empty())
	    {
	      const struct passwd * const pw=getpwnam(u.c_str());
	      if(pw)
		{
		  p.erase(pos, u.size()+1);
		  p.insert(pos, pw->pw_dir);
		}
	      else
		pos=std::string::npos; // we could not find the name, we abort ~ substitution
	    }
	  else
	    {
	      std::string P;
	      if(getenv("HOME"))
		P=getenv("HOME");
	      if(P.empty())
		{
		  const char * const s=getenv("USER");
		  if(s)
		    {
		      const struct passwd * const pw=getpwnam(s);
		      P=pw->pw_dir;
		    }
		}

	      p.erase(pos, 1);
	      p.insert(pos, cleanupPath(P));
	    }
	}
    }
#endif

  // replace // with /
  pos=0;
  while(pos!=std::string::npos)
    {
      pos=p.find(DIRSEP_S DIRSEP_S);
      if(pos!=std::string::npos)
	p.erase(pos,1);
    }

  // eval ..
  pos=0;
  while(pos!=std::string::npos)
    {
      pos=p.find(DIRSEP_S ".." DIRSEP_S);
      if(pos!=std::string::npos)
	{
	  p.erase(pos,3);
	  if(pos>0)
	    {
	      unsigned z=1;
	      while(pos>0 && p[--pos]!=DIRSEP_C) ++z;
	      p.erase(pos,z);
	    }
	}
    }

  // remove .
  pos=0;
  while(pos!=std::string::npos)
    {
      pos=p.find(DIRSEP_S "." DIRSEP_S);
      if(pos!=std::string::npos)
	p.erase(pos,2);
    }

  // remove a trailing ..
  if(p.size()>=3 && (pos=p.find(DIRSEP_S "..")) == p.size()-3)
    {
      p.erase(pos, 3);
      if(pos>0)
	{
	  unsigned z=0;
	  while(pos>0 && p[--pos]!=DIRSEP_C) ++z;
	  if(z)
	    p.erase(pos+1,z);
	}
    }

  // remove a trailing .
  if(p.size()>=2 && (pos=p.find(DIRSEP_S ".")) == p.size()-2) p.erase(pos, 2);

  // remove a leading ..
  if(p.find(".." DIRSEP_S) == 0) p.erase(0, 3);

  // remove a leading .
  if(p.find("." DIRSEP_S) == 0) p.erase(0, 2);

  // remove a trailing /
  if(p.size()>1 && p[p.size()-1]==DIRSEP_C) p.erase(p.size()-1, 1);

  // handle a solely ..
  if(p=="..") p.clear();
  // handle a solely .
  if(p==".") p.clear();

  return p;
}

std::string doj::absFilename(std::string fn)
{
  if(fn.empty())
    return fn;

  char buf[PATH_MAX];
#if defined(__unix__)
  const char * const f=realpath(fn.c_str(), buf);
  if(f)
    return f;
#if 1
  else
    std::cerr << "could not realpath(" << fn << "): " << strerror(errno) << std::endl;
#endif
#endif

  // if relative path?
  if(fn[0]!=DIRSEP_C)
    {
      if(
#if defined(__unix__)
	 getcwd(buf, sizeof(buf))
#elif defined(_MSC_VER)
	 _getdcwd(0, buf, sizeof(buf))
#else
#error unknown platform
#endif
	 )
	fn=std::string(buf)+DIRSEP_S+fn;
    }
  return cleanupPath(fn);
}

std::string doj::dirname(std::string fn)
{
  const std::string::size_type pos=fn.find_last_of(DIRSEP_C);
  if(pos!=std::string::npos)
    fn.erase(pos);
  if(fn.empty())
    fn=DIRSEP_S;
  return fn;
}

std::string doj::filename(std::string fn)
{
  const std::string::size_type pos=fn.find_last_of(DIRSEP_C);
  if(pos!=std::string::npos)
    fn.erase(0, pos+1);
  return fn;
}

#ifdef _MSC_VER
bool doj::exists(const std::string& filename)
{
  struct _stat buffer;
  if(_stat(filename.c_str(), &buffer) < 0)
    return false;
  return true;
}

bool doj::isFile(const std::string& filename)
{
  struct _stat buffer;
  if(_stat(filename.c_str(), &buffer) < 0)
    return false;
  return buffer.st_mode & _S_IFREG;
}
bool doj::isDir(const std::string& filename)
{
  struct _stat buffer;
  if(_stat(filename.c_str(), &buffer) < 0)
    return false;
  return buffer.st_mode & _S_IFDIR;
}

doj::off_t doj::fileSize(const std::string& filename)
{
  struct _stati64 filestat;
  if(_stati64(filename.c_str(), &filestat) == -1)
    return -1;
  return (filestat.st_mode & _S_IFREG) ? filestat.st_size : -1;
}

#define DOJ_TIME(name, record)				\
  time_t name (const std::string& filename) {		\
    struct _stat filestat;				\
    if(_stat(filename.c_str(), &filestat) == -1)	\
      return -1;					\
    return filestat.record;				\
  }

#else
#define DOJ_ISTYPE(name, func) \
  bool name (const std::string& fn) { \
    struct stat filestat; \
    if(lstat(fn.c_str(), &filestat) == -1) \
      return false; \
    return (func(filestat.st_mode))?true:false; \
  }

namespace doj {
  //lint -save -e123 disable error about S_ISREG, S_ISDIR, ... usage
  DOJ_ISTYPE(isFile,S_ISREG)
  DOJ_ISTYPE(isDir,S_ISDIR)
  DOJ_ISTYPE(isCharDev,S_ISCHR)
  DOJ_ISTYPE(isBlockDev,S_ISBLK)
  DOJ_ISTYPE(isFifo,S_ISFIFO)
  DOJ_ISTYPE(isLink,S_ISLNK)
  DOJ_ISTYPE(isSocket,S_ISSOCK)
  //lint -restore
}

#undef DOJ_ISTYPE

bool doj::exists(const std::string& fn)
{
  struct stat filestat;
  if(stat(fn.c_str(), &filestat) == -1)
    return false;
  return true;
}

bool doj::isExec(const std::string& fn)
{
  struct stat filestat;
  if(stat(fn.c_str(), &filestat) == -1)
    return false;
  return (S_ISREG(filestat.st_mode) && (filestat.st_mode&(S_IXUSR|S_IXGRP|S_IXOTH)))?true:false;
}

doj::off_t doj::fileSize(const std::string& fn)
{
  struct stat filestat;
  if(lstat(fn.c_str(), &filestat) == -1)
    return -1;
  return (S_ISREG(filestat.st_mode)) ? filestat.st_size : -1;
}

#define DOJ_TIME(name, record) \
  time_t name (const std::string& fn) { \
    struct stat filestat; \
    if(lstat(fn.c_str(), &filestat) == -1) \
      return -1; \
    return filestat.record; \
  }

#endif

namespace doj {
  DOJ_TIME(aTime, st_atime);
  DOJ_TIME(cTime, st_ctime);
  DOJ_TIME(mTime, st_mtime);
}

#undef DOJ_TIME

namespace doj {
  namespace path {
    int hex2dec(const char c)
    {
      switch(c)
	{
	case '0': return 0;
	case '1': return 1;
	case '2': return 2;
	case '3': return 3;
	case '4': return 4;
	case '5': return 5;
	case '6': return 6;
	case '7': return 7;
	case '8': return 8;
	case '9': return 9;
	case 'A':
	case 'a': return 10;
	case 'B':
	case 'b': return 11;
	case 'C':
	case 'c': return 12;
	case 'D':
	case 'd': return 13;
	case 'E':
	case 'e': return 14;
	case 'F':
	case 'f': return 15;
	default: break;
	}
      return -1;
    }
  }
}

std::string doj::url_unescape(const std::string& u)
{
  std::string s; s.reserve(u.size());
  //lint --e{953}
  for(std::string::const_iterator i=u.begin(), e=u.end(); i!=e; ++i)
    if(*i == '%')
      {
	// check first char
	if(++i != e)
	  {
	    const char first=*i;
	    if(isxdigit(first))
	      {
		// check second char
		if(++i != e)
		  {
		    const char second=*i;
		    if(isxdigit(second))
		      {
			s+=static_cast<char>(doj::path::hex2dec(first)*16 + doj::path::hex2dec(second));
		      }
		    else
		      {
			s+='%';
			s+=first;
			s+=second;
		      }
		  }
		else
		  {
		    s+='%';
		    s+=first;
		  }
	      }
	    else
	      {
		s+='%';
		s+=first;
	      }
	  }
	else
	  s+='%';
      }
    else if(*i == '+')
      s+=' ';
    else
      s+=*i;
  return s;
}

std::string doj::url_escape(const std::string& u)
{
  std::string s; s.reserve(u.size());
  //lint --e{953}
  for(std::string::const_iterator i=u.begin(), e=u.end(); i!=e; ++i)
    {
      char buf[sizeof(int)*2+1];
      const char c=*i;
      // warning: one of the ugliest switch blocks following
      switch(c)
	{
	case ' ': s+='+'; break;
	default:
	  if(c<' ' || c>'~')
	    ; // fall through
	  else
	    {
	      s+=c;
	      break;
	    }
	  /*FALLTHRU*/
	case '&':
	case '<':
	case '>':
	case '"':
	case '#':
	case '$':
	case '%':
	case '\'':
	case '+':
	case ',':
	case '/':
	case ':':
	case ';':
	case '=':
	case '?':
	case '@':
	case '[':
	case '\\':
	case ']':
	case '^':
	case '`':
	case '{':
	case '|':
	case '}':
	case '~':
	  //lint -e{534} ignore return value
	  snprintf(buf, sizeof(buf), "%%%02X", static_cast<int>(static_cast<unsigned char>(c)));
	  s+=buf;
	}
    }
  return s;
}

bool doj::isValidFD(int fd)
{
  if(fd<0) return false;
  const int d=dup(fd);
  if(d<0) return false;
  close(d);
  return true;
}

std::string doj::slurpFile(const std::string& fn)
{
  std::string s;

  errno=0;
  const doj::off_t filesize=doj::fileSize(fn);
  if(filesize <= 0)
    {
      syslog(LOG_NOTICE, "doj::slurpFile(%s): filesize=%i\n", fn.c_str(), static_cast<int>(filesize));
      return s;
    }
  if(filesize > 
#if defined(__unix__)
    std::numeric_limits<int>::max()
#elif defined(_MSC_VER)
    INT_MAX
#else
#error unknown platform
#endif
    )
    {
      syslog(LOG_NOTICE, "doj::slurpFile(%s): filesize=%lli too large\n", fn.c_str(), static_cast<long long int>(filesize));
      return s;
    }

  FILE *f=fopen(fn.c_str(), "rb");
  if(!f)
    {
      syslog(LOG_ERR, "doj::slurpFile(%s): could not open: %s\n", fn.c_str(), strerror(errno));
      return s;
    }

  s.resize(static_cast<unsigned>(filesize));

  if(::readn(fileno(f), const_cast<char*>(reinterpret_cast<const char*>(s.data())), static_cast<unsigned>(filesize)) != filesize)
    {
      syslog(LOG_ERR, "doj::slurpFile(%s): could not read: %s\n", fn.c_str(), strerror(errno));
      s.clear();
    }

  fclose(f);
  return s;
}

std::string doj::slurpFile(FILE *f)
{
  std::string s;
  while(!feof(f))
    {
      char buf[4096];
      //lint -e{713} ignore possible loss of precision in return type
      const size_t r=fread(buf, 1, sizeof(buf), f);
      if(r>0)
	s.append(buf, buf+r);
    }
  return s;
}

bool doj::mkdir(const std::string& dn_, const unsigned mode)
{
  const std::string dn=cleanupPath(dn_);

  // check arguments
  if(dn.empty()) return false;
  if(isDir(dn)) return true;
  if(dn == DIRSEP_S) return false;

  // try to create
  if(::mkdir(dn.c_str()
#if defined(__unix__)
	  , mode
#endif
	  ) == 0) return true;

  // it failed, maybe parent directories are missing
  std::string pdir=dn;
  const std::string::size_type pos=pdir.find_last_of('/');
  if(pos == std::string::npos) return false;
  pdir.erase(pos);
  if(!doj::mkdir(pdir, mode)) return false;

  // try again
  if(::mkdir(dn.c_str()
#if defined(__unix__)
	  , mode
#endif
	  ) == 0) return true;

  // no success, this directory can not be created
  return false;
}

doj::filesInDir_t doj::filesInDir(const std::string& dn)
{
  //lint --e{954,953} variables are ok the non const way they a are defined here

  filesInDir_t s;
#if defined(__unix__)
  DIR *dir=opendir(dn.c_str());
  if(dir)
    {
      struct dirent *d;
      while((d=readdir(dir)) != 0)
	{
	  const std::string fn=d->d_name;
	  if(fn=="." || fn=="..") continue;
	  const std::string fn_=dn+DIRSEP_S+fn;
	  if(isFile(fn_))
	    {
	      //lint -e{534} ignore return value
	      s.insert(fn_);
	    }
	}
      if(closedir(dir) < 0)
	syslog(LOG_ERR, "doj::filesInDir(%s): could not closedir: %s\n", dn.c_str(), strerror(errno));
    }
#elif defined(_MSC_VER)
#warning filesInDir not implemented!
  // \todo write this function with findfirst(), findnext(), findclose()
#endif
  return s;
}


#ifdef TESTPATH
#include <assert.h>
#include <iostream>
#include "readn.c"
using namespace doj;

void a(const std::string& a, const std::string& b, int line)
{
  if(a==b) return;
  std::cout << __FILE__ << ':' << line << ":\"" << a << "\"!=\"" << b << "\"\n";
  exit(1);
}
#define aa(l,r) a(l,r,__LINE__)
int main()
{
  assert(doj::mkdir("/tmp/__sfd___/a/b/c/d/e/f/g/g/h"));
  assert(isDir("/tmp/__sfd___/a/b/c/d/e/f/g/g/h"));

  system("touch /tmp/__sfd___/dirk");
  system("touch /tmp/__sfd___/oliver");
  system("touch /tmp/__sfd___/jagdmann");
  system("mkfifo /tmp/__sfd___/mist");

  doj::filesInDir_t fid=doj::filesInDir("/tmp/__sfd___");
  assert(fid.size() == 3);

  system("rm -rf /tmp/__sfd___");

  assert(isValidFD(0));
  assert(isValidFD(1));
  assert(isValidFD(2));
  assert(!isValidFD(3));
  assert(!isValidFD(-234));
  assert(!isValidFD(0x7FFFFFF));

  aa(url_escape(""), "");
  aa(url_escape("a"), "a");
  aa(url_escape("/"), "%2F");
  aa(url_escape("%"), "%25");
  aa(url_escape("|"), "%7C");
  aa(url_escape(" "), "+");
  aa(url_escape(""), "%C4");
  aa(url_escape(" <>#%{}|\\^~[]`;/?:@=&doj$"), "+%3C%3E%23%25%7B%7D%7C%5C%5E%7E%5B%5D%60%3B%2F%3F%3A%40%3D%26doj%24");
  aa(url_escape("dirk ist hier"), "dirk+ist+hier");

  aa(url_unescape(""), "");
  aa(url_unescape("a"), "a");
  aa(url_unescape("%2F"), "/");
  aa(url_unescape("%25"), "%");
  aa(url_unescape("%7C"), "|");
  aa(url_unescape("%20"), " ");
  aa(url_unescape("%C4"), "");
  aa(url_unescape("%ZZ"), "%ZZ");
  aa(url_unescape("dirk+%ZZ%20%tg"), "dirk %ZZ %tg");
  aa(url_unescape("%20%3C%3E%23%25%7B%7D%7C%5C%5E%7E%5B%5D%60%3B%2F%3F%3A%40%3D%26doj%24"), " <>#%{}|\\^~[]`;/?:@=&doj$");
  aa(url_unescape("dirk+ist+hier"), "dirk ist hier");

  aa(cleanupPath("/bla/../doj"), "/doj");
  aa(cleanupPath("./doj"), "doj");
  aa(cleanupPath("../../../../"), "");
  aa(cleanupPath(".."), "");
  aa(cleanupPath("../"), "");
  aa(cleanupPath("../d//."), "d");
  aa(cleanupPath("/../"), "/");
  aa(cleanupPath("./"), "");
  aa(cleanupPath("./a//../bla/doj/.././g"), "bla/g");
  aa(cleanupPath("a/b/.."), "a");
  aa(cleanupPath("~"), getenv("HOME"));
#ifdef __APPLE__
  aa(cleanupPath("~root/.."), "/var");
#else
  aa(cleanupPath("~root/.."), "/");
#endif
  aa(cleanupPath(""), "");

  aa(dirname(""), "/");
  aa(dirname("/"), "/");
  aa(dirname("a"), "a");
  aa(dirname("a/"), "a");
  aa(dirname("/tmp/"), "/tmp");
  aa(dirname("/tmp"), "/");
  aa(dirname("/tmp/a"), "/tmp");
  aa(dirname(".."), "..");
  aa(dirname("../a"), "..");

  aa(filename(""), "");
  aa(filename("a"), "a");
  aa(filename("a b"), "a b");
  aa(filename("/bla/"), "");
  aa(filename("/bla/z"), "z");
  aa(filename("/bla/z/file.txt"), "file.txt");

#ifdef __APPLE__
  aa(absFilename("/etc/passwd"), "/private/etc/passwd");
  aa(absFilename("/etc/../etc/passwd"), "/private/etc/passwd");
#else
  aa(absFilename("/etc/passwd"), "/etc/passwd");
  aa(absFilename("/etc/../etc/passwd"), "/etc/passwd");
#endif
  aa(absFilename("/"), "/");
  aa(absFilename(""), "");

  assert(exists("/etc/passwd"));
  assert(!exists("/etc/DOJpasswd"));
  assert(isFile("/etc/passwd"));
  assert(!isFile("/tmp"));
  assert(!isDir("/etc/passwd"));
  assert(isDir("/tmp") || isLink("/tmp"));
  assert(isLink("/dev/stdin"));
  assert(!isLink("/dev/null"));
  assert(isCharDev("/dev/zero"));
  assert(!isCharDev("/tmp"));
#ifdef __linux__
  assert(isBlockDev("/dev/sda"));
#endif
  assert(!isBlockDev("/tmp"));
  assert(isExec("/bin/sh"));
  assert(!isExec("/etc/passwd"));
  assert(!isExec("/tmp"));

  assert(fileSize("/etc/passwd") > 0);
  assert(fileSize("/tmp") < 0);
  assert(fileSize("gibt's nich") < 0);

  return 0;
}
#endif

//lint -restore
